#include "Agent.h"
#include "Agent.h"

const float SizeScale = 1.0 / 1000.0;

FAgent::FAgent(FBuildingTile* Home, FVector Offset, FAgentSettings& Settings)
{
	Position = Home->GetLocation();
	WorldOffset = Offset;

	Transform.SetTranslation(Position + WorldOffset);
	Velocity = FVector(0);
	Acceleration = FVector(0);

	HomeTile = Home;
	FoodTarget = nullptr;

	Size = Settings.Size;
	Radius = Settings.Radius;
	PersonalSpaceEpsilon = Settings.PersonalSpace;
	PushingLambda = Settings.Pushing;
	WalkTolerance = Settings.WalkCheckpointTolerance;
	HungerLevel = Settings.HungerLevel;

	FoodCapacity = Settings.FoodCapacity;
	FoodHungerDelta = Settings.FoodHungerDelta;
	FoodEatingDelta = Settings.FoodEatingDelta;
	PatienceLimit = Settings.PatienceLimit;
	PatienceDelta = Settings.PatienceDelta;
	TalkingDistance = Settings.TalkingDistance;

	WeightWallRepulsion = Settings.WeightWallRepulsion;
	WeightAgentRepulsion = Settings.WeightAgentRepulsion;
	WeightWallAvoidance = Settings.WeightWallAvoidance;
	WeightAgentAvoidance = Settings.WeightAgentAvoidance;
	WeightSteeringForce = Settings.WeightSteeringForce;

	Patience = PatienceLimit;
	Food = FoodCapacity / 10;
	StopCounter = 0;

	Status = EAgentState::Home;

	Transform.SetScale3D(FVector(Size * SizeScale)); // scale the ball down
}

FAgent::~FAgent()
{
}

void FAgent::UpdateStatus(FCityMap* Map, TArray<FAgent*>& Others)
{
	if (Status != EAgentState::Eating)
		Food = std::max(0.0f, Food - FoodHungerDelta);

	switch (Status)
	{
		case Home:
			if (Food < HungerLevel)
				if (MoveToStreet(Map))
					GoSearchForFood(Map);
			break;
		case Searching:
			if (WalkAction.EndReached())
				GoToFoodmarketDoor(Map);
			break;
		case Waiting:
			if (IsGoingToWait(Map))
				if (WalkAction.EndReached()) //try walk towards door
					if (TryGetInsideFoodmarket(Map))
						StartEating(Map);
					else
						WaitAtFoodmarketDoor(Map);
			break;
		case WalkingAround:
			if (!NewMarketDiscovered(Map, Others))
				if (WalkAction.EndReached())
					PlanRandomWalk(Map);
			//else if new is discovered, goes to searching
			break;
		case Eating:
			if (Food < FoodCapacity)
				Eat(Map);
			else if (MoveToStreet(Map))
				GoHome(Map);
			break;
		case Returning:
			if (WalkAction.EndReached())
				GoToHomeDoor(Map);
			break;
		case GoingToHomeDoor:
			if (WalkAction.EndReached())
				StayHome(Map);
			break;
		default:
			break;
	}

	UpdateWaitInterval();
}

bool FAgent::MoveToStreet(FCityMap* Map)
{
	FBuildingTile* Curtile = static_cast<FBuildingTile*>(Map->GetTileForLocation(Position));
	FTile* Street = Map->GetNearestTile(Map->GetTileForLocation(Position), ETileType::Street);

	if (Street)
	{
		FVector Door = Curtile->DoorLocation(Street);
		Position = Door;
		return true;
	}

	return false;
}


bool FAgent::GoSearchForFood(FCityMap* Map)
{
	FoodTarget = static_cast<FFoodTile*>(Map->GetNearestTile(Map->GetTileForLocation(Position), ETileType::Food));
	if (!FoodTarget)
	{
		return false;
	}

	FTile* StreetInFrontFood = Map->GetNearestTile(FoodTarget, ETileType::Street);
	if (!StreetInFrontFood)
	{
		return false;
	}

	WalkAction = FWalk(Position, StreetInFrontFood->GetLocation(), Map, WalkTolerance);
	Status = EAgentState::Searching;

	return true;
}

bool FAgent::GoToFoodmarketDoor(FCityMap* Map)
{

	FTile* Current = Map->GetTileForLocation(Position);
	FVector Inside = FoodTarget->DoorLocation(Current);
	WalkAction = FWalk(Position, Inside, Map, WalkTolerance);
	Status = EAgentState::Waiting;
	return true;
}

bool FAgent::TryGetInsideFoodmarket(FCityMap* Map)
{
	return FoodTarget->TryAgentGetInside();
}


bool FAgent::WaitAtFoodmarketDoor(FCityMap* Map)
{
	FTile* Current = Map->GetTileForLocation(Position);
	FVector Inside = FoodTarget->DoorLocation(Current);
	WalkAction = FWalk(Position, Inside, WalkTolerance);
	return true;
}

bool FAgent::IsGoingToWait(FCityMap* Map)
{
	Patience = std::max(0.0f, Patience - PatienceDelta);

	if (Patience < 0.5)
	{
		PlanRandomWalk(Map);
		Patience = PatienceLimit;
		return false;
	}

	return true;
}

bool FAgent::PlanRandomWalk(FCityMap* Map)
{
	Status = EAgentState::WalkingAround;
	FTile* Tile = Map->GetRandomStreetTile();
	WalkAction = FWalk(Position, Tile->GetLocation(), Map, WalkTolerance);

	return true;
}

bool FAgent::NewMarketDiscovered(FCityMap* Map, TArray<FAgent*>& Others)
{
	for (const auto& o : Others)
	{
		if (o->IsStopped() || Dist(o) > TalkingDistance)
			continue;

		if (o->Food != Food)
		{
			Food = o->Food;
			FTile* StreetInFrontFood = Map->GetNearestTile(FoodTarget, ETileType::Street);
			if (!StreetInFrontFood)
			{
				return false;
			}

			WalkAction = FWalk(Position, StreetInFrontFood->GetLocation(), Map, WalkTolerance);
			Status = EAgentState::Searching;

			StopCounter = Map->GetRandomWaitInterval();
			o->StopCounter = StopCounter;

			return true;
		}
	}

	return false;
}

bool FAgent::StartEating(FCityMap* Map)
{
	Position = FoodTarget->GetInsideAgentPosition(this);
	Status = EAgentState::Eating;
	return true;
}

bool FAgent::Eat(FCityMap* Map)
{
	Food += FoodEatingDelta;
	return true;
}

bool FAgent::GoHome(FCityMap* Map)
{
	FoodTarget->AgentGetOut(this);
	FTile* StreetInFrontHome = Map->GetNearestTile(HomeTile, ETileType::Street);
	WalkAction = FWalk(Position, StreetInFrontHome->GetLocation(), Map, WalkTolerance);
	Status = EAgentState::Returning;
	return true;
}

bool FAgent::GoToHomeDoor(FCityMap* Map)
{
	FTile* Current = Map->GetTileForLocation(Position);
	FVector Inside = HomeTile->DoorLocation(Current);
	WalkAction = FWalk(Position, Inside, Map, WalkTolerance);
	Status = EAgentState::GoingToHomeDoor;
	return true;
}

bool FAgent::StayHome(FCityMap* Map)
{
	Position = HomeTile->GetLocation();
	Status = EAgentState::Home;
	return true;
}

FVector FAgent::OtherAgentsRepulsionForce(TArray<FAgent*>& Others) const {
	
	FVector Force = FVector(0);
	// TODO: Implement the force which creates repulsion between agents.
	// This force is based on the personal space of the agent and it repels them
	// from every other agent.
	// See Agent.h for available variables. Start with PersonalSpaceEpsilon, Radius and Position.

	return Force;
}


FVector FAgent::OtherAgentsAvoidanceForce(TArray<FAgent*>& Others) const {

	FVector Force = FVector(0);
	// TODO: Implement the force which makes the actors avoid each other.
	// This force helps avoid collisions. You need to consider every other agent.
	// See Agent.h for available variables. Start with Position, Radius and PersonalSpaceEpsilon.

	return Force;
}

void FAgent::ForcesCompute(FCityMap* Map, TArray<FAgent*>& Others)
{
	UpdateStatus(Map, Others);

	//avoidances
	AvoidanceWallForce = Map->WallsAvoidanceForce(Position, Velocity, Radius, PersonalSpaceEpsilon);
	AvoidanceAgentsForce = OtherAgentsAvoidanceForce(Others);

	//repulsions
	RepulsionWallForce = Map->WallsRepulsionForce(Position, Radius, PersonalSpaceEpsilon);
	RepulsionAgentsForce = OtherAgentsRepulsionForce(Others).GetSafeNormal();
}

void FAgent::CheckStoppingRule()
{
	Alpha = (Repulsion.SizeSquared() > 0) && (FVector::DotProduct(Velocity, RepulsionAgentsForce) < 0) ? 0 : 1;
}

void FAgent::UpdateWaitInterval()
{
	StopCounter = std::max(0, StopCounter - 1);
}

float FAgent::Dist(FAgent* Agent) const
{
	return (Position - Agent->Position).Size();
}

int FAgent::IsNotLookingForOthers() const
{
	return Status == EAgentState::WalkingAround ? 0 : 1;
}

bool FAgent::IsStopped() const
{
	return StopCounter > 0;
}

void FAgent::ForcesApply(float DeltaTime)
{
	float Z = Position.Z;

	//apply forces
	if (WalkAction.GetForce(Position, AttractorForce) == EWalkStatus::Walking)
	{
		Steering = WeightSteeringForce * AttractorForce
			+ WeightWallAvoidance * AvoidanceWallForce
			+ WeightAgentAvoidance * AvoidanceAgentsForce * IsNotLookingForOthers();

		Steering = Steering.GetSafeNormal();

		Repulsion = RepulsionWallForce * WeightWallRepulsion
			+ WeightAgentRepulsion * PushingLambda * RepulsionAgentsForce * (0.5 + 0.5 * IsNotLookingForOthers());

		Repulsion = Repulsion.GetSafeNormal();

		CheckStoppingRule(); // setting alpha

		Acceleration = Alpha * Steering + Repulsion;

		Velocity += Acceleration.GetClampedToMaxSize(1.0);
		Velocity = Velocity.GetClampedToMaxSize(1.0) * std::min(DeltaTime, 1.0f); //to prevent large steps
		Position += Velocity;
		Position.Z = Z; //to fix floating
	}

	Transform.SetTranslation(Position + WorldOffset);
}

FVector FAgent::GetPosition() const
{
	return Position;
}

float FAgent::GetFeatureState() const
{
	if (StopCounter)
		return 7.1;

	switch (Status)
	{
	case Home:
		return 0.1;
	case Searching:
		return 1.1;
	case Waiting:
		return 2.1;
	case WalkingAround:
		return 3.1;
	case Eating:
		return 4.1;
	case Returning:
		return 5.1;
	case GoingToHomeDoor:
		return 6.1;
	default:
		break;
	}

	return 0.1;
}