#include "CityGrid.h"
#include "CitySettings.h"
#include "ProceduralMeshComponent.h"
#include "Components/InstancedStaticMeshComponent.h"

// Sets default values
ACityGrid::ACityGrid()
{
 	// Set this actor to call Tick() every frame.  You can turn this off to improve performance if you don't need it.
	PrimaryActorTick.bCanEverTick = true;

	// Generate base geometry	
	Mesh = CreateDefaultSubobject<UProceduralMeshComponent>(TEXT("CityMesh"));
	static ConstructorHelpers::FObjectFinder<UMaterial> MaterialOb(TEXT("Material'/Game/CrowdSim/VertexMaterial.VertexMaterial'"));
	Material = MaterialOb.Object;
	// All transformations are applied to this mesh
	RootComponent = Mesh;

	//generate city
	Layout = nullptr;
	GenerateMesh = false;
	Speed = 2;
	Height = 20;
	Spacing = 20;
	DisplayCustomersInRow = 3;
	LimitCustomersInside = 9;

	//individual
	PersonalSpaceRangeMin = 1.0;
	PersonalSpaceRangeMax = 5.0;
	PushFactorRangeMin = 0.0;
	PushFactorRangeMax = 2.0;
	FoodHungerDeltaMin = 0.1;
	FoodHungerDeltaMax = 0.1;
	FoodEatingDeltaMin = 1.0;
	FoodEatingDeltaMax = 1.0;
	FoodCapacityMin = 100.0;
	FoodCapacityMax = 100.0;
	PatienceLimitMin = 1000.0;
	PatienceLimitMax = 1000.0;
	PatienceDeltaMin = 0.5;
	PatienceDeltaMax = 0.5;

	//common
	PerceptionRadius = 15;
	WalkCheckpointTolerance = 4;
	AgentRadius = 1;
	TalkingDistance = 4;
	WeightWallRepulsion = 1.0;
	WeightAgentRepulsion = 1.0;
	WeightWallAvoidance = 1.0;
	WeightAgentAvoidance = 1.0;
	WeightSteeringForce = 10.0;
	HungerLevel = 5;

	//initiate agents mesh
	AgentMesh = CreateDefaultSubobject<UInstancedStaticMeshComponent>(TEXT("AgentMesh")); // A component that efficiently renders multiple instances of the same StaticMesh.
	AgentMesh->NumCustomDataFloats = 1; // Defines the number of floats that will be available per instance for custom data
	AgentMesh->RegisterComponent();

	static ConstructorHelpers::FObjectFinder<UMaterial> MaterialAgentOb(TEXT("Material'/Game/CrowdSim/AgentMaterial.AgentMaterial'"));
	UStaticMesh* agentMeshInstance = ConstructorHelpers::FObjectFinder<UStaticMesh>(TEXT("StaticMesh'/Engine/BasicShapes/Sphere.Sphere'")).Object;

	agentMeshInstance->SetMaterial(0, MaterialAgentOb.Object);

	AgentMesh->SetStaticMesh(agentMeshInstance);

	Map = nullptr;
}

// Called when the game starts or when spawned
void ACityGrid::BeginPlay()
{
	Super::BeginPlay();

	InitCity();
	InitAgents();
}

// Called every frame
void ACityGrid::Tick(float DeltaTime)
{
	Super::Tick(DeltaTime);

	TArray<FAgent*> Others;
	
	//update agents, compute forces
	for (size_t i = 0; i < Agents.Num(); ++i)
	{
		Map->GetOccupantsForPosition(Others, Agents[i]->GetPosition(), PerceptionRadius);
		Agents[i]->ForcesCompute(Map, Others);
	}
	
	//reset occupants
	Map->EmptyOccupants();
	

	FVector Loc;
	FTile* Tile;

	//update agent geometry position
	for (size_t i = 0; i < Agents.Num(); ++i)
	{
		Agents[i]->ForcesApply(DeltaTime * Speed * 10.0f);
		AgentMesh->UpdateInstanceTransform(i, Agents[i]->Transform, true, i == (Agents.Num() - 1));
		AgentMesh->SetCustomDataValue(i, 0, Agents[i]->GetFeatureState(), i == (Agents.Num() - 1));

		Loc = Agents[i]->GetPosition();
		Tile = Map->GetTileForLocation(Loc);

		//to prevent all evil
		if (Tile)
			Tile->AddOccupant(Agents[i]);
		else
		{
			Agents.RemoveAt(i);
			i--;
		}
	}
}

void ClampMinMax(float& Min, float& Max)
{
	Min = std::min(Min, Max);
	Max = std::max(Min, Max);
}

void ACityGrid::OnConstruction(const FTransform& Transform)
{
	if (GenerateMesh && Layout)
		InitCity();

	//clamp parameters
	ClampMinMax(PersonalSpaceRangeMin, PersonalSpaceRangeMax);
	ClampMinMax(PushFactorRangeMin, PushFactorRangeMax);
	ClampMinMax(FoodHungerDeltaMin, FoodHungerDeltaMax);
	ClampMinMax(FoodEatingDeltaMin, FoodEatingDeltaMax);
	ClampMinMax(FoodCapacityMin, FoodCapacityMax);
	ClampMinMax(PatienceLimitMin, PatienceLimitMax);
	ClampMinMax(PatienceDeltaMin, PatienceDeltaMax);
}

void ACityGrid::InitCity()
{
	GenerateMesh = false;
	if (Map)
		delete Map;

	FCitySettings Settings;
	Settings.Spacing = Spacing;
	Settings.Height = Height;
	Settings.Material = Material;
	Settings.Layout = Layout;
	Settings.Mesh = Mesh;
	Settings.FoodmarketAgentsRow = DisplayCustomersInRow;
	Settings.FoodmarketLimit = LimitCustomersInside;

	// create CityMap
	Map = new FCityMap();
	Map->Generate(Settings);
}

void ACityGrid::InitAgents()
{
	// common parameters for agents
	FAgentSettings Settings; 
	Settings.Size = Spacing;
	Settings.Radius = AgentRadius;
	Settings.HungerLevel = HungerLevel;
	Settings.WalkCheckpointTolerance = WalkCheckpointTolerance;
	Settings.TalkingDistance = TalkingDistance;
	Settings.WeightWallRepulsion = WeightWallRepulsion;
	Settings.WeightAgentRepulsion = WeightAgentRepulsion;
	Settings.WeightWallAvoidance = WeightWallAvoidance;
	Settings.WeightAgentAvoidance = WeightAgentAvoidance;
	Settings.WeightSteeringForce = WeightSteeringForce;

	// random generator helpers for individual parameters
	std::default_random_engine Generator;
	std::uniform_real_distribution<float> PersonalSpace(PersonalSpaceRangeMin, PersonalSpaceRangeMax);
	std::uniform_real_distribution<float> Push(PushFactorRangeMin, PushFactorRangeMax);
	std::uniform_real_distribution<float> Limit(FoodCapacityMin, FoodCapacityMax);
	std::uniform_real_distribution<float> Hunger(FoodHungerDeltaMin, FoodHungerDeltaMax);
	std::uniform_real_distribution<float> Eat(FoodEatingDeltaMin, FoodEatingDeltaMax);
	std::uniform_real_distribution<float> PatienceLimit(PatienceLimitMin, PatienceLimitMax);
	std::uniform_real_distribution<float> PatienceDelta(PatienceDeltaMin, PatienceDeltaMax);

	// individual parameters for agents
	const TArray<FTile*>& Buildings = Map->GetBuildings();

	FAgent* Agent;
	for (size_t i = 0; i < Buildings.Num(); ++i)
	{
		Settings.PersonalSpace = PersonalSpace(Generator);
		Settings.Pushing = Push(Generator);
		Settings.FoodCapacity = Limit(Generator);
		Settings.FoodHungerDelta = Hunger(Generator);
		Settings.FoodEatingDelta = Eat(Generator);
		Settings.PatienceLimit = PatienceLimit(Generator);
		Settings.PatienceDelta = PatienceDelta(Generator);

		Agent = new FAgent(static_cast<FBuildingTile*>(Buildings[i]), GetTransform().GetLocation(), Settings); // Agent::Agent(BuildingTile* home, FVector woffset, AgentSettings& settings)
		AgentMesh->AddInstance(Agent->Transform);
		Agents.Push(Agent);
	}
}


