// Fill out your copyright notice in the Description page of Project Settings.


#include "CityMap.h"

#include "Tile.h"
#include "BuildingTile.h"
#include "BorderTile.h"
#include "StreetTile.h"
#include "FoodTile.h"

FCityMap::FCityMap()
{
	RandomWaitInterval = std::uniform_int_distribution<int>(100, 500);
}

FCityMap::~FCityMap()
{
	for (size_t i = 0; i < Res; ++i)
		delete Tiles[i];
}

const TArray<FTile*>& FCityMap::GetBuildings() const
{
	return Buildings;
}


FTile* FCityMap::GetTileForLocation(const FVector Loc) const
{
	int32 X = Loc.X / Spacing;
	int32 Y = Loc.Y / Spacing;

	size_t Idx = Y * ResolutionX + X;

	if (Idx >= Res || Loc.X < 0 || Loc.Y < 0)
		return nullptr;

	return Tiles[Idx];
}

FTile* FCityMap::GetRandomStreetTile()
{
	int Idx = RandomStreet(Generator);
	return Streets[Idx];
}

int FCityMap::GetRandomWaitInterval()
{
	return RandomWaitInterval(Generator);
}

void FCityMap::Generate(const FCitySettings& Settings)
{
	Spacing = Settings.Spacing;
	Height = Settings.Height;
	GenerateCity(Settings);
	GenerateMesh(Settings);
}

void FCityMap::GenerateCity(const FCitySettings& Settings)
{
	TextureCompressionSettings OldCompressionSettings = Settings.Layout->CompressionSettings;
	TextureMipGenSettings OldMipGenSettings = Settings.Layout->MipGenSettings;
	bool OldSRGB = Settings.Layout->SRGB;

	Settings.Layout->CompressionSettings = TextureCompressionSettings::TC_VectorDisplacementmap;
	Settings.Layout->MipGenSettings = TextureMipGenSettings::TMGS_NoMipmaps;
	Settings.Layout->SRGB = false;
	Settings.Layout->UpdateResource();

	//prepare texture
	const FColor* FormatedImageData = static_cast<const FColor*>(Settings.Layout->GetPlatformData()->Mips[0].BulkData.LockReadOnly());

	ResolutionX = Settings.Layout->GetPlatformData()->Mips[0].SizeX;
	ResolutionY = Settings.Layout->GetPlatformData()->Mips[0].SizeY;

	for (int32 y = 0; y < ResolutionY + 2; y++)
	{
		for (int32 x = 0; x < ResolutionX + 2; x++)
		{
			if (x == 0 || y == 0 || x == (ResolutionX + 1) || y == (ResolutionY + 1))
			{
				Tiles.Push(new FBorderTile(x, y, Spacing, Spacing, 1));
				continue;
			}

			FColor PixelColor = FormatedImageData[(y - 1) * ResolutionX + (x - 1)];
			float R = PixelColor.R;
			float G = PixelColor.G;

			if (R != G)
			{
				Tiles.Push(new FFoodTile(x, y, Spacing, Spacing, R / 255.f * Height, Settings.FoodmarketLimit, Settings.FoodmarketAgentsRow));
				Food.Push(Tiles.Last());
			}
			else if (R == 0)
			{
				Tiles.Push(new FStreetTile(x, y, Spacing, Spacing));
				Streets.Push(Tiles.Last());
			}
			else
			{
				Tiles.Push(new FBuildingTile(x, y, Spacing, Spacing, R / 255.f * Height));
				Buildings.Push(Tiles.Last());
			}

		}
	}

	ResolutionX += 2;
	ResolutionY += 2;
	Res = (ResolutionX) * (ResolutionY);

	Settings.Layout->GetPlatformData()->Mips[0].BulkData.Unlock();
	Settings.Layout->CompressionSettings = OldCompressionSettings;
	Settings.Layout->MipGenSettings = OldMipGenSettings;
	Settings.Layout->SRGB = OldSRGB;
	Settings.Layout->UpdateResource();

	RandomStreet = std::uniform_int_distribution<int>(0, Streets.Num() - 1);

	for (size_t i = 0; i < Res; ++i)
		Tiles[i]->SetupNeighbors(Tiles, ResolutionX, ResolutionY);
}

void FCityMap::GenerateMesh(const FCitySettings& Settings)
{
	//mesh arrays
	TArray<FVector> Vertices;
	TArray<FVector> Normals;
	TArray<int32> Indices;
	TArray<FVector2D> UVs;
	TArray<FLinearColor> VertexColors;
	TArray<FProcMeshTangent> Tangents;
	int VertexCount = 0;

	for (size_t i = 0; i < Res; ++i)
		VertexCount += Tiles[i]->GenerateMesh(Vertices, Normals, Indices, UVs, VertexCount, VertexColors);

	Settings.Mesh->CreateMeshSection_LinearColor(0, Vertices, Indices, Normals, UVs, VertexColors, Tangents, false);
	Settings.Mesh->SetMaterial(0, Settings.Material);
}

//Street or Food
FTile* FCityMap::GetNearestTile(FTile* Tile, ETileType Type) const
{
	TQueue<FTile*> Queue;
	TSet<FTile*> Visited;
	Queue.Enqueue(Tile);
	Visited.Add(Tile);

	while (!Queue.IsEmpty())
	{
		Queue.Dequeue(Tile);

		if (Tile->GetTileType() != Type)
		{
			const TArray<FTile*>& Neighbors = Tile->GetNeighbors();
			for (const auto& Neighbor : Neighbors)
			{
				if (!Visited.Contains(Neighbor))
				{
					Queue.Enqueue(Neighbor);
					Visited.Add(Neighbor);
				}
			}
		}
		else {
			return Tile;
		}
	}

	return nullptr;
}

void FCityMap::EmptyOccupants()
{
	for (size_t i = 0; i < Res; ++i)
		Tiles[i]->EmptyOccupants();
}

void FCityMap::GetOccupantsForPosition(TArray<FAgent*>& Occupants, const FVector& Pos, float Radius)
{
	TQueue<FTile*> Queue;
	TSet<FTile*> Visited;

	float Radius2 = Radius * Radius;
	FTile* Tile = GetTileForLocation(Pos);
	Queue.Enqueue(Tile);
	Visited.Add(Tile);

	while (!Queue.IsEmpty())
	{
		Queue.Dequeue(Tile);
		Tile->GetOccupants(Occupants);

		const TArray<FTile*>& Neighbors = Tile->GetNeighbors();

		for (const auto& Neighbor : Neighbors)
		{
			if (Neighbor->GetVecToClosestTilePoint(Pos).SizeSquared() < Radius2)
				if (!Visited.Contains(Neighbor))
				{
					Queue.Enqueue(Neighbor);
					Visited.Add(Neighbor);
				}
		}
	}
}

FVector FCityMap::WallsAvoidanceForce(const FVector& Pos, const FVector& Dir, const float AgentRadius, const float Epsilon)
{
	FTile* Tile = GetTileForLocation(Pos);

	const TArray<FTile*>& Neighbors = Tile->GetNeighbors();
	FVector Force = FVector(0), WallDir, Normal;
	float Dist, DistAttr;

	for (const auto& Neighbor : Neighbors)
	{		
		if (Neighbor->GetTileType() == ETileType::Street)
			continue;

		WallDir = Neighbor->GetVecToClosestTilePoint(Pos);

		Dist = WallDir.Size();
		Normal = Neighbor->NearestWallNormal(Pos);

		DistAttr = AgentRadius + Epsilon - Dist;

		if (DistAttr > 0)
			Force += FVector::CrossProduct(FVector::CrossProduct(Normal, Dir), Normal).GetSafeNormal();
	}

	return Force;
}

FVector FCityMap::WallsRepulsionForce(const FVector& Pos, const float AgentRadius, const float Epsilon)
{
	FTile* Tile = GetTileForLocation(Pos);

	const TArray<FTile*>& Neighbors = Tile->GetNeighbors();

	FVector WallDir, Normal;
	FVector Force = FVector(0);
	float Dist, DistAttr;

	for (const auto& Neighbor : Neighbors)
	{
		if (Neighbor->GetTileType() == ETileType::Street)
			continue;

		WallDir = Neighbor->GetVecToClosestTilePoint(Pos);

		Dist = WallDir.Size();
		Normal = Neighbor->NearestWallNormal(Pos);

		DistAttr = AgentRadius + Epsilon - Dist;
		if (DistAttr > 0)
			Force += (Normal * DistAttr) / Dist;
	}

	return Force;
}