// Fill out your copyright notice in the Description page of Project Settings.

#include "Walk.h"
#include "CityMap.h"

FWalk::FWalk() {}

bool Collinear(FVector A, FVector B, FVector C) {
	return fabs((A.Y - B.Y) * (A.X - C.X) - (A.Y - C.Y) * (A.X - B.X)) <= 1e-9;
}

FWalk::FWalk(const FVector Start, const FVector End, const FCityMap* Map, const float Tol)
{
	Tolerance = Tol * Tol;
	TMap<const FTile*, const FTile*> Pred;

	const FTile* StartTile = Map->GetTileForLocation(Start);
	const FTile* EndTile = Map->GetTileForLocation(End);

	//bfs to find trajectory
	BFS(Pred, StartTile, EndTile);
	auto Backtrack = Pred.Find(EndTile);

	if (!Backtrack)
	{
		return; // nothing to be done, there is not path
	}

	Checkpoints.Push(End);

	FVector Origin = End;
	FVector Checkpoint;

	//setup trajectory checkpoints
	while (*Backtrack != nullptr)
	{
		Checkpoint = (*Backtrack)->GetLocation();

		if (Checkpoints.Num() == 1)
		{
			Checkpoints.Push(Checkpoint);
		}
		else if (Collinear(Origin, Checkpoints.Last(), Checkpoint))
		{
			Checkpoints.Last() = Checkpoint;
		}
		else {
			Origin = Checkpoints.Last();
			Checkpoints.Push(Checkpoint);
		}

		Backtrack = Pred.Find(*Backtrack);
	}
}

FWalk::FWalk(const FVector Start, const FVector End, const float _tolerance)
{
	Checkpoints.Push(End);
	Checkpoints.Push(Start);
	Tolerance = _tolerance * _tolerance;
}

void FWalk::BFS(TMap<const FTile*, const FTile*>& Pred,
	const FTile* StartTile,
	const FTile* EndTile)
{
	TQueue<const FTile*> Queue;
	const FTile* Tile = StartTile;
	Queue.Enqueue(Tile);
	Pred.Add(Tile, nullptr);

	while (!Queue.IsEmpty())
	{
		Queue.Dequeue(Tile);

		const TArray<FTile*>& Neighbors = Tile->GetNeighbors();

		for (const auto Neighbor : Neighbors)
		{
			if (Neighbor->GetTileType() != ETileType::Street)
				continue;

			if (!Pred.Find(Neighbor))
			{
				Pred.Add(Neighbor, Tile);
				Queue.Enqueue(Neighbor);

				if (EndTile == Neighbor)
				{
					//path found
					return;
				}
			}
		}
	}
}

FWalk::~FWalk()
{
}

int FWalk::GetForce(const FVector AgentLoc, FVector& Force)
{
	if (Checkpoints.Num() == 0)
	{
		Force = FVector(0);
		return EWalkStatus::Finished;
	}

	//move to next checkpoint when the agent 
	//is in a certain distance from checkpoint
	//tolerance distance set to 2
	FVector Dir = Checkpoints.Last() - AgentLoc;
	if (Dir.SizeSquared() < Tolerance)
	{
		Checkpoints.Pop();

		if (Checkpoints.Num() == 0)
		{
			Force = FVector(0);
			return EWalkStatus::Finished;
		}

		Dir = Checkpoints.Last() - AgentLoc;
	}

	Force = Dir.GetSafeNormal();
	return EWalkStatus::Walking;
}

bool FWalk::EndReached() {
	return Checkpoints.Num() == 0;
}